package ru.dbotthepony.mc.otm.graph.storage

import net.minecraftforge.common.util.LazyOptional
import ru.dbotthepony.mc.otm.graph.Graph6Node
import ru.dbotthepony.mc.otm.storage.IStorageIdentity
import ru.dbotthepony.mc.otm.storage.IStorageStack
import java.util.*

open class BasicStorageGraphNode : IStorageGraphNode {
	private var resolver = LazyOptional.of<IStorageGraphNode> { this }
	private var valid = true
	protected val components = ArrayList<IStorageIdentity<*>>()

	private val node = Graph6Node<IStorageGraphNode>(this)

	override fun getAsStorageNode(): Graph6Node<IStorageGraphNode> {
		return node
	}

	override fun getComponents(): List<IStorageIdentity<*>> {
		return Collections.unmodifiableList(components)
	}

	fun <T : IStorageStack, U : IStorageIdentity<T>> computeIfAbsent(identity: Class<T>, provider: () -> U): U {
		for (component in components) {
			if (component.storageIdentity() == identity) {
				return component as U
			}
		}

		val factory = provider()
		addStorageComponent(factory)
		return factory
	}

	fun <T : IStorageStack, U : IStorageIdentity<T>> computeIfAbsent(identity: Class<T>, provider: (Class<T>) -> U): U {
		for (component in components) {
			if (component.storageIdentity() == identity) {
				return component as U
			}
		}

		val factory = provider(identity)
		addStorageComponent(factory)
		return factory
	}

	fun addStorageComponent(component: IStorageIdentity<*>) {
		for (component1 in components) {
			if (component === component1 || component1.storageIdentity() == component.storageIdentity()) {
				return
			}
		}

		components.add(component)
		getStorageGraph()?.add(component)
	}

	fun removeStorageComponent(component: IStorageIdentity<*>) {
		for (component1 in components) {
			if (component === component1 || component1.storageIdentity() == component.storageIdentity()) {
				components.remove(component1)
				getStorageGraph()?.remove(component)
				return
			}
		}
	}

	fun invalidate() {
		if (!valid) return
		valid = false
		resolver.invalidate()
		// detachStorage();
	}

	fun revive() {
		if (valid) return
		valid = true
		resolver = LazyOptional.of { this }
		// attachStorage();
	}

	fun get(): LazyOptional<IStorageGraphNode> {
		return if (valid) resolver else LazyOptional.empty()
	}
}