package ru.dbotthepony.mc.otm.menu

import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.*
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.block.entity.BlockEntity
import ru.dbotthepony.mc.otm.menu.widget.AbstractWidget
import java.util.function.Consumer

@JvmRecord
data class MoveResult(val mergeOccurred: Boolean, val remaining: ItemStack, val changed_slots: Set<Slot>)

abstract class MatteryMenu protected @JvmOverloads constructor(
	p_38851_: MenuType<*>?,
	p_38852_: Int,
	@JvmField val inventory: Inventory,
	@JvmField val tile: BlockEntity? = null
) : AbstractContainerMenu(p_38851_, p_38852_) {
	@JvmField
	val ply: Player = inventory.player

	@JvmField
	val matteryWidgets = ArrayList<AbstractWidget>()

	@JvmField
	val inventorySlots = ArrayList<MatterySlot>()

	@JvmField
	protected val lockedInventorySlots: MutableSet<Int> = HashSet()
	protected open fun isInventorySlotLocked(index: Int): Boolean = lockedInventorySlots.contains(index)

	@JvmField
	protected var _synchronizer: ContainerSynchronizer? = null

	override fun setSynchronizer(p_150417_: ContainerSynchronizer) {
		_synchronizer = p_150417_
		super.setSynchronizer(p_150417_)
	}

	fun addWidget(widget: AbstractWidget, consumer: Consumer<Consumer<ContainerData>>) {
		if (!matteryWidgets.contains(widget)) {
			matteryWidgets.add(widget)
			consumer.accept(Consumer { p_38885_: ContainerData -> addDataSlots(p_38885_) })
		}
	}

	@JvmField
	protected var inventorySlotIndexStart = 0

	@JvmField
	protected var inventorySlotIndexEnd = 0

	@JvmOverloads
	protected fun addInventorySlots(offset: Int = 97) {
		var first = true

		for (i in 0..2) {
			for (j in 0..8) {
				val slot: MatterySlot = object : MatterySlot(inventory, j + i * 9 + 9, 8 + j * 18, 14 + i * 18) {
					override fun mayPlace(p_40231_: ItemStack): Boolean {
						return !isInventorySlotLocked(index)
					}

					override fun mayPickup(p_40228_: Player): Boolean {
						return !isInventorySlotLocked(index)
					}
				}

				inventorySlots.add(slot)
				addSlot(slot)

				if (first) {
					first = false
					inventorySlotIndexStart = slot.index
				}
			}
		}

		var last: MatterySlot? = null

		for (k in 0..8) {
			last = object : MatterySlot(inventory, k, 8 + k * 18, 14 + 58) {
				override fun mayPlace(p_40231_: ItemStack): Boolean {
					return !isInventorySlotLocked(index)
				}

				override fun mayPickup(p_40228_: Player): Boolean {
					return !isInventorySlotLocked(index)
				}
			}

			addSlot(last)
			inventorySlots.add(last)
		}

		inventorySlotIndexEnd = last!!.index
	}

	override fun broadcastChanges() {
		for (widget in matteryWidgets) {
			widget.updateServer()
		}

		super.broadcastChanges()
	}

	override fun broadcastFullState() {
		for (widget in matteryWidgets) {
			widget.updateServer()
		}

		super.broadcastFullState()
	}

	override fun stillValid(player: Player): Boolean {
		if (tile == null) return true

		if (player.level.getBlockEntity(tile.blockPos) !== tile) {
			return false
		}

		val pos = tile.blockPos

		return player.distanceToSqr(pos.x.toDouble() + 0.5, pos.y.toDouble() + 0.5, pos.z.toDouble() + 0.5) <= 64.0
	}

	protected abstract fun getWorkingSlotStart(): Int
	protected abstract fun getWorkingSlotEnd(): Int

	// This method receive Player interactor and slot_index where Shift + Right click occurred
	// It shall return item stack that got moved
	override fun quickMoveStack(ply: Player, slot_index: Int): ItemStack {
		// this.moveItemStackTo(ItemStack, int start_slot_index, int end_slot_index, boolean iteration_order)
		// returns boolean, telling whenever ItemStack was modified (moved or shrank)
		// false means nothing happened
		// Last boolean determine order of slot iteration, where:
		// if TRUE, iteration order is end_slot_index -> start_slot_index
		// if FALSE iteration order is start_slot_index -> end_slot_index
		val start = getWorkingSlotStart()
		val end = getWorkingSlotEnd()

		if (start == end) {
			return ItemStack.EMPTY
		}

		var moved = ItemStack.EMPTY
		val initialSlot = slots[slot_index]

		if (initialSlot.hasItem()) {
			val initialItem = initialSlot.item
			moved = initialItem.copy()

			if (slot_index < inventorySlotIndexStart) {
				// Moving FROM machine TO inventory
				if (!moveItemStackTo(initialItem, inventorySlotIndexStart, inventorySlotIndexEnd + 1, false)) {
					return ItemStack.EMPTY
				}
			} else if (!moveItemStackTo(initialItem, start, end, false)) {
				// Moving FROM inventory TO machine
				return ItemStack.EMPTY
			}

			if (initialItem.isEmpty) {
				initialSlot.set(ItemStack.EMPTY)
			} else {
				initialSlot.setChanged()
			}
		}

		return moved
	}

	fun quickMoveToInventory(stack: ItemStack, simulate: Boolean): MoveResult {
		return if (inventorySlotIndexStart == 0 && inventorySlotIndexEnd == 0) {
			MoveResult(false, stack, setOf())
		} else customMoveItemStackTo(
			stack,
			inventorySlotIndexStart,
			inventorySlotIndexEnd + 1,
			false,
			simulate
		)
	}

	override fun moveItemStackTo(
		stack_to_move: ItemStack,
		initial_slot: Int,
		final_slot: Int,
		reverse_direction: Boolean
	): Boolean {
		val moveResult = customMoveItemStackTo(stack_to_move, initial_slot, final_slot, reverse_direction, false)
		val remaining: ItemStack = moveResult.remaining
		if (remaining.count == stack_to_move.count) return false
		stack_to_move.count = remaining.count
		return moveResult.mergeOccurred
	}

	protected fun customMoveItemStackTo(
		stackToMove1: ItemStack,
		initial_slot: Int,
		final_slot: Int,
		reverse_direction: Boolean,
		simulate: Boolean
	): MoveResult {
		var mergeOccurred = false
		var i = if (reverse_direction) final_slot - 1 else initial_slot
		val changed = HashSet<Slot>()
		val stackToMove = stackToMove1.copy()

		if (stackToMove.isStackable) {
			while (!stackToMove.isEmpty) {
				if (reverse_direction) {
					if (i < initial_slot) {
						break
					}
				} else if (i >= final_slot) {
					break
				}

				val slot = slots[i]
				val slot_stack = slot.item

				if (!slot_stack.isEmpty && ItemStack.isSameItemSameTags(stackToMove, slot_stack)) {
					val j = slot_stack.count + stackToMove.count
					val maxSize = slot.getMaxStackSize(stackToMove)

					if (j <= maxSize) {
						stackToMove.count = 0

						if (!simulate) {
							slot_stack.count = j
							slot.setChanged()
						}

						mergeOccurred = true
						changed.add(slot)
					} else if (slot_stack.count < maxSize) {
						stackToMove.shrink(maxSize - slot_stack.count)

						if (!simulate) {
							slot_stack.count = maxSize
							slot.setChanged()
						}

						mergeOccurred = true
						changed.add(slot)
					}
				}

				i += if (reverse_direction) -1 else 1
			}
		}

		if (!stackToMove.isEmpty) {
			i = if (reverse_direction) final_slot - 1 else initial_slot

			while (true) {
				if (reverse_direction) {
					if (i < initial_slot) {
						break
					}
				} else if (i >= final_slot) {
					break
				}

				val slot = slots[i]
				val slot_stack = slot.item

				if (slot_stack.isEmpty && slot.mayPlace(stackToMove)) {
					if (!simulate) {
						if (stackToMove.count > slot.maxStackSize) {
							slot.set(stackToMove.split(slot.maxStackSize))
						} else {
							slot.set(stackToMove.split(stackToMove.count))
						}

						slot.setChanged()
					}

					mergeOccurred = true
					changed.add(slot)
					break
				}

				i += if (reverse_direction) -1 else 1
			}
		}

		return MoveResult(mergeOccurred, stackToMove, changed)
	}
}